package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)BaseNode.java 1.21 98/11/05 20:34:02
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

public abstract class BaseNode {
	String defName;
	Browser browser;
	Loader loader;
	boolean implReady=false;
	javax.media.j3d.SceneGraphObject implObject;
	BaseNode parent;
	static final boolean debugBaseNode=false;
	static String v97Pkg = new String("com.sun.j3d.loaders.vrml97.impl.");

	BaseNode(Loader loader) { 
	this.loader = loader;
	this.browser = loader.browser;
	if (debugBaseNode)
		System.out.println(toStringId()+" created with browser "+browser.hashCode());
	//System.out.println("Node " + this + " created");
	}
    public abstract Object clone();
    void define(String defName) {
        this.defName = defName;
        //System.out.println("Node " + this + " def'd to name " + defName);
    }
    public Browser getBrowser() { return browser; }
    public abstract Field getField(String fieldName);

	/* Note that the implNode is actually just the root of the graph for this object, and not
	 * necessarily connected with the interesting mechanics of this object.  E.g. Transform objects.
	 */
	public javax.media.j3d.Node getImplNode() {
	javax.media.j3d.SceneGraphObject sgo = getImplObject();
	if (sgo instanceof javax.media.j3d.Node) {
	    return (javax.media.j3d.Node) sgo;
	} else {
	    return null;
	}
	}
    public javax.media.j3d.SceneGraphObject getImplObject() {
        return(implObject);
    }
    // subclasses which have triangles should define
    public int getNumTris() {
        return(0);
    }
    public abstract String getType();
    public void initImpl() { 
	if (loader.debug)  {
	    System.out.println("BaseNode.initImpl() called on node " + 
		toStringId());
	}
	implObject = null; 
	implReady = true; 
    }
    public abstract void notifyMethod(String s, double time);
    void registerUse(Scene scene) {;}
    // subclasses should implement toStringBody() not toString()
    public final String toString() {
        String retval = "";
        if( defName != null )
        {
            retval += "DEF " + defName + " ";
        }
        retval += toStringBody();
        return(retval);
    }
    // this should be overridden by a method which outputs the VRML for the
    // node
    String toStringBody() {
        return(toStringId());
    }
    // use instead of toString() to get object id instead of VRML source
    public String toStringId() {
	String className = this.getClass().getName();
	if (className.startsWith(v97Pkg)) {
	    className = className.substring(v97Pkg.length());
	}
	return className + "@" + 
		Integer.toHexString(this.hashCode());
    }
    // called by parent after node is added to parent
    void updateParent(javax.media.j3d.Node parentImpl) {;}
/**
 * Inform the node that the scene graph above it has changed.
 * TimeSensors and other nodes the Browser object should know about should
 * respond by performing the appropriate registrations.
 *
 * Technically, a node passed between browsers should also update its ownership.
 */
public void updateSceneGraph(BaseNode newParent) {
	parent=newParent;
	if (parent!=null) {
		browser=parent.browser;
		if (debugBaseNode)
			if (browser==null)
				System.out.println(toStringId()+" acquiring a NULL browser from "+newParent.toStringId()+".");
			else
				System.out.println(toStringId()+" acquiring browser "+browser.hashCode()+" from "+newParent.toStringId()+".");
	} else {
		browser=null;
		if (debugBaseNode)
			System.out.println(toStringId()+" loosing browser reference.");
	}
}
    public abstract vrml.BaseNode wrap();
    // Script is the funny one. If there are any other
    // cases, for now ignore by this method. Node will 
    // implement fully
    void writeX3D(java.io.Writer w, int depth) throws java.io.IOException {;}
    // give VrmlScene a hook into the impl writer.
    public void writeX3D(java.io.Writer w, com.sun.j3d.loaders.vrml97.VrmlScene vs) 
		throws java.io.IOException {
        writeX3D(w,1);
    }
}
