package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)Billboard.java 1.21 99/03/15 10:27:43
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

import javax.media.j3d.Group;
import javax.media.j3d.TransformGroup;

public class Billboard extends GroupBase {

	// exposedField 
	SFVec3f axisOfRotation;
	// Note: this field does *not* change as the Billboard rotates, rather,
	// this is the axis about which it rotates.

	Group impl;
	TransformGroup implTrans;
	javax.media.j3d.Billboard implBillboard;

	
    public Billboard(Loader loader) {
	super(loader);
	axisOfRotation = new SFVec3f(0.0f,1.0f,0.0f);
	initBillboardFields();
    }
    Billboard(Loader loader, MFNode children, SFVec3f bboxCenter,
	    SFVec3f  bboxSize,  SFVec3f aor ) {
	super(loader, children, bboxCenter, bboxSize);
	axisOfRotation = aor;
	initBillboardFields();
    }
    public Object clone() {
	return new Billboard(loader, 
		(MFNode)children.clone(),
	        (SFVec3f)bboxCenter.clone(),
	        (SFVec3f)bboxSize.clone(),
		(SFVec3f)axisOfRotation.clone() );
    }
	public javax.media.j3d.Group getGroup() {
		return implTrans;
	}
    public String getType() { return "Billboard"; }
    void initBillboardFields() {
	axisOfRotation.init(this, FieldSpec, Field.EXPOSED_FIELD,
	    "axisOfRotation");
    }
    void initFields() {
	super.initFields();
	initBillboardFields();
    }
	public void initImpl() { 
		impl = new javax.media.j3d.BranchGroup();
		implTrans = new TransformGroup();
		implNode = impl;
		implTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		implTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		impl.setCapability(javax.media.j3d.BranchGroup.ALLOW_DETACH);
		implTrans.setCapability(Group.ALLOW_CHILDREN_WRITE);
		implTrans.setCapability(Group.ALLOW_CHILDREN_READ);
	
		implBillboard = new javax.media.j3d.Billboard(implTrans);
		setAxis();
		impl.addChild(implBillboard);
		impl.addChild(implTrans);
		implBillboard.setTarget(implTrans);
		implBillboard.setSchedulingBounds(loader.infiniteBounds);
		super.replaceChildren(); // add the children to the group
		implReady = true;
	}
    public void notifyMethod(String eventInName, double time) {
	if ( eventInName.equals("axisOfRotation")) {
	    setAxis();
	} else if ( eventInName.equals("route_axisOfRotation")) {
	    ; // No-op
	} else {
	    super.notifyMethod(eventInName, time);
	}
    }
    private void setAxis() {
	float[] axis = axisOfRotation.value;
	if ((axis[0] == 0.0) && (axis[1] == 0.0) &&
	    (axis[2] == 0.0)) {
	    implBillboard.setAlignmentMode(
			javax.media.j3d.Billboard.ROTATE_ABOUT_POINT);
	    implBillboard.setRotationPoint(0.0f, 0.0f, 0.0f);
	} else {
	    implBillboard.setAlignmentMode(
			javax.media.j3d.Billboard.ROTATE_ABOUT_AXIS);
	    implBillboard.setAlignmentAxis(axis[0],axis[1],axis[2]);
	}
    }
}
