package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)BindableNode.java 1.18 98/11/05 20:34:04
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

import java.util.Stack;

public abstract class BindableNode extends Node {

	// eventIn
	SFBool bind;     
	SFTime bindTime; 
  
	// eventOut
	SFBool isBound;     

	Stack stack;	

    BindableNode(Loader loader, Stack initStack) { 
	super(loader);
	stack = initStack;
	bind = new SFBool(true);
	bindTime = new SFTime();
	isBound = new SFBool(false);
    }
    BindableNode(Loader loader, Stack initStack, SFBool bind,
		SFTime bindTime, SFBool isBound) { 
	super(loader);
	stack = initStack;
	this.bind = bind;
	this.bindTime = bindTime;
	this.isBound = isBound;
    }
    public javax.media.j3d.SceneGraphObject getImplObject() {
	return implObject;
    }
	void initBindableFields() {
	bind.init(    this, FieldSpec, Field.EVENT_IN,  "set_bind");
	bindTime.init(this, FieldSpec, Field.EVENT_OUT, "bindTime");
	isBound.init( this, FieldSpec, Field.EVENT_OUT, "isBound");
	}
public void notifyMethod(String eventInName, double time) {
	if (browser==null)
	  return;
	BindableNode top;
	if (eventInName.equals("set_bind")) {
		System.out.println("Stack size before set bind:" + stack.size());
		if (bind.getValue()) {
			System.out.println(getClass().getName() + " bind for " + hashCode());
			if (stack.empty()) {
				System.out.println("Stack is empty.");
				top = null;
			} else {
				top = (BindableNode) stack.peek();
			}
			if (top != this) {
				System.out.println("Placing a new element on stack top.");
				if (stack.contains(this)) {
					System.out.println("Removing self.");
					stack.removeElement(this);
				}
				System.out.println("Adding to top of stack.");
				stack.push(this); // no matter what
				browser.bindableChanged(stack);
				bindTime.setValue(time);
			}
			isBound.setValue(true);
		} else { // bind == false
			System.out.println(getClass().getName() + " unbound for " + hashCode());
			if (stack.empty()) {
				System.out.println("Stack is empty.");
				// Deliberately do nothing.
			} else {
				if ((BindableNode) stack.peek() == this) { // i'm on top
					System.out.println("The bound node was top of stack.");
					stack.pop(); // remove this from stack
					// TODO: timestamp of both setValues should be same
					if (!stack.empty()) {
						System.out.println("Binding the new stack top.");
						top = (BindableNode) stack.peek();
						if (top != null) {
						   	// If the top doesn't already have bind of true, something is wrong.
							top.isBound.setValue(true);
						}
					} else {
						System.err.println(
							getClass().getName()
								+ " stack had only one element in it when the top was unbound.");
					}
					browser.bindableChanged(stack);
				} else {
					if (stack.contains(this)) {
						stack.removeElement(this);
					}
				}
			}
			isBound.setValue(false);
		}
		System.out.println("Stack size after set_bind:" + stack.size());
	}
}
}
