package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)CoordinateInterpolator.java 1.22 98/11/05 20:34:22
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

public class CoordinateInterpolator extends Interpolator {

	// eventIn
	SFFloat fraction;
	// exposedField
	// MFFloat key; // From Interpolator
	MFVec3f keyValue;
	// eventOut
	MFVec3f value;
	
    public CoordinateInterpolator(Loader loader) {
	super(loader);
	fraction = new SFFloat(0.0f);

	//there should be empty constructors for all fields.
	key = new MFFloat( );
	keyValue = new MFVec3f();
	value = new MFVec3f();

	initFields();

    }
    public CoordinateInterpolator(Loader loader, SFFloat fraction,
	    MFFloat key, MFVec3f keyValue ) {
	super(loader);

	this.fraction = fraction;
	this.key = key;
	this.keyValue = keyValue;
	this.value = new MFVec3f();
	initFields();
    }
    public Object clone() { 
	return new CoordinateInterpolator(loader, (SFFloat)fraction.clone(),
		(MFFloat)key.clone(), (MFVec3f)keyValue.clone() );
    }
    public String getType() { return "CoordinateInterpolator"; }
	void initFields() {
	fraction.init(this, FieldSpec, Field.EVENT_IN,      "set_fraction");
	key.init(     this, FieldSpec, Field.EXPOSED_FIELD, "key");
	keyValue.init(this, FieldSpec, Field.EXPOSED_FIELD, "keyValue");
	value.init(   this, FieldSpec, Field.EVENT_OUT,     "value_changed");
	}
	public void notifyMethod(String eventInName, double time) {
	if ( eventInName.equals("set_fraction") ) {
	    if (key.mfloat.length > 0) {
		setIndexFract(fraction.value);
		int valsPerKey = keyValue.size / key.mfloat.length;
		if (value.size != valsPerKey) {
		    value.checkSize(valsPerKey, false);
		    value.size = valsPerKey;
		}
		int indL = iL * valsPerKey ;
		int indH = iH * valsPerKey ;
		int v1xi,v1yi,v1zi,v2xi,v2yi,v2zi;
		float v1x,v1y,v1z,v2x,v2y,v2z;
		float vals[] = keyValue.value;
		for (int j = 0; j < valsPerKey  ; j+=3) {
		    // since were stack based, maybe this
		    // can be that much more optimal, ie
		    // inc, pop, push, inc -> inc, inc 
		    v1xi = (indL) + j;  
		    v1yi = v1xi+1;  
		    v1zi = v1yi+1;  
		    //v2xi = v1xi+valsPerKey;
		    v2xi = (indH) + j;  
		    v2yi = v2xi+1;  
		    v2zi = v2yi+1;  
		    v1x = vals[v1xi];
		    v1y = vals[v1yi];
		    v1z = vals[v1zi];
		    v2x = vals[v2xi];
		    v2y = vals[v2yi];
		    v2z = vals[v2zi];
		    value.value[j + 0] = (v1x*af) + (v2x*f);
		    value.value[j + 1] = (v1y*af) + (v2y*f);
		    value.value[j + 2] = (v1z*af) + (v2z*f);
		}
	    }
	    value.route();
	}
	    
	}
}
