package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)Field.java 1.25 99/03/09 16:36:15
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

/* Added implements clause */
public abstract class Field implements FieldEventReceiver {
	String fieldName = new String("?");
	BaseNode ownerNode = null;
	double lastUpdate = -1.0; // to allow one update only in loader case
	Vector connections = null;
	public ConstField constField;
	public int fieldType = EXPOSED_FIELD; // least restrictive for transition
	boolean hidden=false;

	// Field types
	static final int    FIELD = 4;
	static final int    EVENT_IN = 1;
	static final int    EVENT_OUT = 2;
	static final int    EXPOSED_FIELD = EVENT_IN | EVENT_OUT | FIELD;

	// debugging
	static final boolean printRoutes = false;

	// remove the "set_" or "_changed" from a fieldName 
	public static String baseName(String fieldName) {
	String newName = fieldName;
	if (fieldName.startsWith("set_")) {
	    newName = fieldName.substring(4);
	    //System.out.println("Field.baseName() Removed set_: old " +
	    //	fieldName + " new: " + newName);
	}
	if (fieldName.endsWith("_changed")) {
	    newName = fieldName.substring(0, fieldName.indexOf("_changed"));
	    //System.out.println("Field.baseName() Removed _changed: old " +
	    //	fieldName + " new: " + newName);
	}
	return newName;
	}
	public abstract Object clone();
	public void connectToField(FieldEventReceiver field) {
		System.out.print("connectToField called on "+this.toStringId()+" by "+field.toStringId()+".\n");
		if( connections == null ) {
			System.out.print("Initializing connections list.\n");
			connections = new Vector();
		}
		if( !connections.contains(field) ) {
			System.out.print("Adding field "+field.toStringId()+" to receiver list.\n");
			connections.addElement(field);
		}
	}
	// should initialize constField
	abstract ConstField constify();
	void deleteConnection(FieldEventReceiver field) {
		connections.removeElement(field);
	}
	public String getFieldName() {
	  return fieldName;
	}
	public double getLastUpdate() {
	  return lastUpdate;
	}
	public abstract String getX3DValue();
// should be called for all fields in Node in initFields()
void init(BaseNode node, Hashtable fieldSpec, int fieldType, String fieldName) {
	this.ownerNode = node;
	this.fieldType = fieldType;
	this.fieldName = fieldName;
	if (fieldSpec != null) {
		fieldSpec.put(fieldName, this);
		if (this.fieldType == Field.EXPOSED_FIELD) {
			fieldSpec.put(fieldName + "_changed", this);
			fieldSpec.put("set_" + fieldName, this);
		}
	}
}
/**
 * hidden == true means that X3D will not print out the field
 */
	void init(BaseNode node, Hashtable fieldSpec, int fieldType,
			  String fieldName, boolean hidden) {
		this.ownerNode = node;
		this.fieldType = fieldType;
		this.fieldName = fieldName;
	// fieldType overrules hidden
	this.hidden=hidden;
	if( fieldType==EVENT_OUT ) {
	    this.hidden=true;
	}
		if( fieldSpec != null ) {
			fieldSpec.put(fieldName, this);
		}
	}
	boolean isEventIn() {
		return((fieldType & EVENT_IN) != 0);
	}
	boolean isEventOut() {
		return((fieldType & EVENT_OUT) != 0);
	}
	// marks nodes connected to this node as being writable.  Used when
	// a PROTO field is marked writable to make the fields it is connected
	// to be writable too.
	public void markConnectedWriteable() {
	if (connections != null) {
	    for (Enumeration e = connections.elements(); 
		    e.hasMoreElements();) {
		Field routeField = (Field)e.nextElement();
		routeField.markWriteable();
	    }
	}
	}
	// This should be done only to nodes that are j3d impls that
	// benefit from compile. Eg, audioclip now has to if check each
	// simtick for route_ prefix but it never uses the prefix.
	// possibly an interface called RoutePrefixable ? If ownernode
	// instance of RoutePrefixable then notifyMethod.
	public void markWriteable() {
		ownerNode.notifyMethod("route_" + fieldName, 0.0d);
	}
	void route() {
		routeUpdate(true);
	}

	// to implement ROUTE
	// Node: fields used to be connected using ConstField->Field.  Changed
	// to do this here so that all the timing logic is together.  ConstFields
	// are now just wrappers that limit access to Fields.
	synchronized void routeUpdate(boolean tellOwner) {
		double eventTime;
		if( ownerNode != null ) {
			if( ownerNode.browser != null ) {
				eventTime = ownerNode.browser.beginRoute();
			} else {
				// if no browser, then time is always 0.0.  lastUpdate is
				// initialized to -1.0 to allow one update but no more.
				eventTime = 0.0;
			}
			if( printRoutes ) {
				System.out.println("Owner of field " + this.toStringId() +
								   " is " + ownerNode.toStringId());
				System.out.println("Field type is " + fieldType);
			}
			if( tellOwner && isEventIn() && ownerNode.implReady ) {
				if( printRoutes ) {
					System.out.println("Calling notifyMethod() on owner");
				}
				ownerNode.notifyMethod(fieldName,eventTime);
			}
		} else {
			eventTime = Time.getNow();
		}
		// Test time to avoid sending two events with same timestamp from
		// a EventOut
		if( lastUpdate != eventTime ) {
			lastUpdate = eventTime;
			if( connections != null ) {
				for( Enumeration e = connections.elements(); e.hasMoreElements(); ) {		
					/* Modified this line */
					FieldEventReceiver routeField = (FieldEventReceiver)e.nextElement();
					// Test time to avoid sending two event with same
					// timestamps to an EventIn
					
					// Changed to be getLastUpdate() instead of field access
					if( routeField.getLastUpdate() != eventTime ) {
						if( printRoutes ) {
							System.out.println("Field.route: field " +
											   fieldName + " " + this + "\n    is updating " +
											   routeField.getFieldName() + " " + routeField);
						}
						routeField.update(this);
					} else {
						if( printRoutes ) {
							System.out.println("Field.route: Not updating " +
								 "routeField " + routeField.getFieldName() + " " +
								 routeField + "\n     lastUpdate (" +
								 ownerNode.browser.relativeTime(routeField.getLastUpdate()) +
									  ") == eventTime");
						}
					}
				}
			}
		} else {
			if( printRoutes ) {
				System.out.println("After update, field " + fieldName + " " +
					 this + "\n     lastUpdate (" +
					 ownerNode.browser.relativeTime(lastUpdate) +") == eventTime, not routing");
			}
		}
		if( (ownerNode != null) && (ownerNode.browser != null) ) {
			ownerNode.browser.endRoute();
		}
	}
	public String toStringId() {
		return(this.getClass().getName() + "@" +
			  Integer.toHexString(this.hashCode()));
	}
	abstract public void update(Field field);
	public abstract vrml.Field wrap();
}
