package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)MFColor.java 1.12 98/11/05 20:34:38
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 * @Author: Auvo Severinkangas
 *
 */

import java.util.Vector;
import java.util.Observer;
import java.util.Observable;

public class MFColor extends MField implements MFNumeric {

	float[] vals; 

    public MFColor( ) {
	float[] dummy = new float[0];
	setValue( dummy );
    }
    public MFColor(float[][] values) {
	setValue(values);
    }
    public MFColor(float[] values) {
	setValue(values);
    }
    // What is the size doing here?
    public MFColor(int size, float[] values) {
	setValue(values);
    }
    public void clear() {
	vals = new float[1];
	route();
    }
    public synchronized Object clone() {
	MFColor ref = new MFColor( );
	ref.vals = new float[ vals.length ];
	try {
	    System.arraycopy( vals,0,ref.vals,0,vals.length );
	} catch (Exception e) {
	    System.err.println("MFColor.clone(): exception " + e);
	}
		
	return ref;
    }
    public synchronized ConstField constify() {
	if (constField == null) {
	    constField = new ConstMFColor(this);
	}
	return constField;
    }
    public void delete(int index) {
	int i;
	float[] temp = new float[vals.length - 3];
	for(i = 0; i < index * 3; i++ ) 
	    temp[i] = vals[i];
	for (; i < temp.length; i++) {
	     temp[i] = vals[i+3];
	}
	vals = temp;
	route();
    }
    public void get1Value(int index, float[] values) {
	System.arraycopy(vals,index * 3,values,0,3);
    }
    public void get1Value(int index, SFColor vec) {
	vec.setValue( vals[index], vals[index+1], vals[index+2] );
    }
    public int getSize() {
	return vals.length / 3;
    }
    public void getValue(float[][] values) {
	int numVecs = vals.length / 3;
	for ( int i = 0 ; i < numVecs; i++ ) {
	    System.arraycopy(vals,i*3,values[i],0,3);
	}
    }
    public void getValue(float[] values) {
	System.arraycopy(vals,0,values,0,vals.length);
    }
    public String getX3DValue()
    {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < getSize(); i+=3) {
	    if(i>1) sb.append(",");
	    if((i%21 == 0) && (i > 0)) sb.append("\n");
	    sb.append(" "+vals[i]+" "+vals[i+1]+" "+vals[i+2]);
        }
        return sb.toString();
    }
    public void insertValue(int index, float x, float y, float z) { 
	try {
	    int i;
	    float[] temp = new float[vals.length + 3];
	    for(i = 0; i < index * 3; i++ ) 
		temp[i] = vals[i];
	    temp[i++] = x;
	    temp[i++] = y;
	    temp[i++] = z;
	    for (; i < temp.length; i++) {
		 temp[i] = vals[i-3];
	    }
	    vals = temp;
	} catch ( IndexOutOfBoundsException e ) { 
	    System.err.println("MFColor.insertValue(index,float,float,float): "+
		"exception " + e);
	} 
	route();
    }
    public void insertValue(int index, ConstSFColor constvec) {
	insertValue(index, (SFColor)constvec.ownerField);
    }
    public void insertValue(int index, SFColor vec) {
	insertValue(index, vec.color[0], vec.color[1], vec.color[2]);
    }
    public void set1Value(int index, float x, float y, float z) {
	try {
	    vals[index*3 + 0] = x;
	    vals[index*3 + 1] = y;
	    vals[index*3 + 2] = z;

	} catch ( IndexOutOfBoundsException e ) { 
	    System.err.println("MFColor.set1Value(index,float,float,float): " +
		"exception " + e);
	}  
	route();
    }
    public void set1Value(int index, ConstSFColor constvec) {
	set1Value(index, (SFColor)constvec.ownerField);
    }
    public void set1Value(int index, SFColor vec) {
	set1Value(index, vec.color[0], vec.color[1], vec.color[2]);
    }
    public void setValue(float[][] values) {
	vals = new float[values.length * 3];
	for ( int i = 0; i < values.length; i++ ) {
	    System.arraycopy(values,0,vals,i*3,i*3 + 3 );
	}
	route();

    }
    public void setValue(float[] values) {
	vals = new float[values.length];
	System.arraycopy(values,0,vals,0,values.length);
	route();
    }
    public void setValue(int size, float[] values) {
	setValue(values);
    }
    public void setValue(ConstMFColor cf) {
	setValue((MFColor)cf.ownerField);
    }
    public void setValue(MFColor color) {
	setValue(color.vals);
    }
    public void update(Field field) {
	setValue((MFColor)field);
    }
    public vrml.Field wrap() {
	return new vrml.field.MFColor(this);
    }
}
