package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)MFNode.java 1.18 98/11/05 20:34:40
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 * @Author: Auvo Severinkangas
 *
 */

import java.util.Observer;
import java.util.Observable;

public class MFNode extends MField {

	BaseNode[] nodes;

    public MFNode() {
	nodes = new BaseNode[0];
    }
    public MFNode(BaseNode[] values) {
	nodes = new BaseNode[values.length];
	System.arraycopy(values,0,nodes,0,values.length);
    }
    public void addValue(BaseNode f) {
	BaseNode[] temp = new BaseNode[nodes.length + 1];
	for ( int i = 0 ; i < temp.length-1; i ++)
	    temp[i] = nodes[i];
	temp[temp.length-1] = f;
	nodes = temp;

	route();
    }
	public void addValue(ConstSFNode constsfnod) {
		if (constsfnod==null)
			addValue((BaseNode)null);
		else
			addValue((SFNode)constsfnod.ownerField);
	}
	public void addValue(SFNode sfnod) {
		if (sfnod==null)
			addValue((BaseNode)null);
		else
			addValue(sfnod.node);
	}
    public void clear() { nodes = null; }
	public synchronized Object clone() {
	MFNode ref = new MFNode(nodes);
	// deep copy 
	for (int i = 0; i < ref.nodes.length; i++) {
	    if (nodes[i] != null) {
		BaseNode cloneNode = (BaseNode)nodes[i].clone();
		ref.nodes[i] = cloneNode;
		if (nodes[i].loader.debug) {
		    System.out.println("MFNode.clone(): child " + i + " " +
			nodes[i].toStringId() + " is cloned to " + 
			cloneNode.toStringId() + " = " + cloneNode);
		}
		nodes[i].loader.registerClone(nodes[i], cloneNode);
	    }
	}
	return ref;
	}
    // Special clone method used by ProtoInstance
    public MFNode cloneMap(ProtoInstance instance) {
	MFNode ref = new MFNode();
	ref.nodes = new BaseNode[nodes.length];
	// shallow copy, mapping old node names to new node names 
	for (int i = 0; i < nodes.length; i++) {
	    if (nodes[i] != null) {
		BaseNode oldNode = (BaseNode)nodes[i];
		BaseNode newNode = instance.updateNode(oldNode);
		if (nodes[i].loader.debug) {
		   System.out.println("MFNode.cloneMap(), map node " +
			oldNode + " to node " + newNode);
		}
		ref.nodes[i] = newNode;
	    }
	}
	return ref;
    }
    public synchronized ConstField constify() {
	if (constField == null) {
	    constField = new ConstMFNode(this);
	}
	return constField;
    }
    // todo
    public void delete(int n) {;}
    public BaseNode get1Value(int index) {
	return ( nodes[index] );
    }
    public int getSize() { return nodes.length; }
    public BaseNode[] getValue() {
	return nodes;
    }
    public void getValue(BaseNode values[]) {
	System.arraycopy(nodes,0,values,0,nodes.length);
    }
    public String getX3DValue() { 
	java.io.StringWriter s = new java.io.StringWriter();
	for (int i = 0; i<nodes.length; i++) {
	    try {
	  	nodes[i].writeX3D(s,1);
	    } catch ( java.io.IOException ioe ) {
		System.out.println("IOEXception writing out x3d for MFNodes "+this);
		System.out.println("Possible Proto IS map error?");
	    }
	} 
	return s.toString();	
    }
    public void insertValue( int index, BaseNode f ) {
	BaseNode[] temp = new BaseNode[nodes.length + 1];
	for ( int i = 0 ; i < index ; i ++ )
		temp[i] = nodes[i];
	temp[index] = f;
	for ( int i = index+1; i < temp.length + 1 ; i++ )
		temp[i] = nodes[i-1];
	nodes = temp;
	route();
    }
    public void insertValue( int index, ConstSFNode constsfnod ) {
	insertValue( index, (SFNode)constsfnod.ownerField );
    }
    public void insertValue( int index, SFNode constsfnod ) {
	insertValue( index, constsfnod.node );
    }
	/* The difference between setValue and replaceValue is that replaceValue doesn't
	 * perform change notification. */
	public void replaceValue(BaseNode[] values) {
	nodes = new BaseNode[ values.length ];
	System.arraycopy(values,0,nodes,0,values.length);
	}
    public void set1Value(int index, BaseNode f) {
	nodes[index] = f;
	route();
    }
    public void set1Value(int index, ConstSFNode csfn) {
	set1Value(index, (SFNode)csfn.ownerField);
    }
    public void set1Value(int index, SFNode sfnod) {
	set1Value(index, sfnod.node);
    }
    public void setValue(BaseNode[] values) {
	nodes = new BaseNode[ values.length ];
	System.arraycopy(values,0,nodes,0,values.length);
	route();
    }
    public void setValue(int size, BaseNode[] value) { 
	// does this really mean take the first size of value?
	setValue(value);
    }
    public void setValue(ConstMFNode constNode) {
	setValue((MFNode)constNode.ownerField);
    }
    public void setValue(MFNode node) {
	setValue((BaseNode[])node.nodes);
    }
    public String toString() {
	String retval = "[\n";
	for (int i = 0; i < nodes.length; i++) {
	    retval += nodes[i] + "\n";
	}
	retval += "]\n";
	return retval;
    }
    public synchronized void update(Field field) {
	setValue((MFNode)field);
    }
    public vrml.Field wrap() {
        return new vrml.field.MFNode(this);
    }
}
