package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)OrientationInterpolator.java 1.18 98/11/05 20:34:50
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

import javax.vecmath.AxisAngle4f;
import javax.vecmath.Quat4f;

public class OrientationInterpolator extends Interpolator {

	// eventIn
	SFFloat fraction;

	// exposedField
	// MFFloat key; // From Interpolator
	MFRotation keyValue;

	// eventOut

	SFRotation value;

	float[] v1,v2;
	// temporaries used for interpolation
	AxisAngle4f a1 =  new AxisAngle4f();
	AxisAngle4f a2 =  new AxisAngle4f();
	Quat4f q1 =  new Quat4f();
	Quat4f q2 =  new Quat4f();
	Quat4f qResult =  new Quat4f();
	AxisAngle4f aResult =  new AxisAngle4f();
	int i,j;

    public OrientationInterpolator(Loader loader) {
	super(loader);
	fraction = new SFFloat(0.0f);
	key = new MFFloat();
	keyValue = new MFRotation();
	value = new SFRotation(0.0f,0.0f,1.0f,0.0f);
	initFields();
    }
    OrientationInterpolator(Loader loader, MFFloat key, MFRotation keyValue) {
	super(loader);
	// initialize the interpolator to the right
	// value. otherwise do it manually;
	this.fraction = new SFFloat(0.0f);
	this.key = key;
	this.keyValue = keyValue;
	this.value = new SFRotation(0.0f,0.0f,1.0f,0.0f);
	initFields();
    }
    public Object clone() {
	return new OrientationInterpolator(loader, (MFFloat)key.clone(),
				(MFRotation)keyValue.clone() );
    }
    public String getType() { return "OrientationInterpolator"; }
	void initFields() {
	fraction.init(this, FieldSpec, Field.EVENT_IN,      "set_fraction");
	key.init(     this, FieldSpec, Field.EXPOSED_FIELD, "key");
	keyValue.init(this, FieldSpec, Field.EXPOSED_FIELD, "keyValue");
	value.init(   this, FieldSpec, Field.EVENT_OUT,     "value_changed");
	}
	public void notifyMethod(String eventInName, double time) {
	if ( eventInName.equals("set_fraction") ) {
	    //System.out.println("OrientationInterpolator: fraction");
	    if (key.mfloat.length > 0) {
		setIndexFract(fraction.value);
		//System.out.println("OrientationInterpolator: " +
		//	"fr = " + fraction.value + " iL = " + iL +
		//	" f = " + f + " af = " + af);
		// Linear interpolation leads to bogus results.  Instead,
		// convert to quaternions and interpolate
		// TODO: could cache quat form of rotations
		a1.set(keyValue.rots[iL].rot);
		a2.set(keyValue.rots[iH].rot);
		q1.set(a1);
		q2.set(a2);
		qResult.interpolate(q1, q2, f);
		aResult.set(qResult);
		//System.out.println("\ta1 = " + a1 + "\n\ta2 = " + a2 +
		//	"\n\taResult = " + aResult);;
		value.rot[0] = aResult.x;
		value.rot[1] = aResult.y;
		value.rot[2] = aResult.z;
		value.rot[3] = aResult.angle;
	    }
	    value.route();
	}

	}
}
