package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)PointSet.java 1.16 98/11/05 20:34:53
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 *
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 */

import javax.media.j3d.PointArray;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.BoundingBox;
import javax.vecmath.Point3d;

public class PointSet extends Geometry {

	PointArray impl; 
	BoundingBox bounds;

	SFNode color;
	SFNode coord;

	int vertexFormat=0;
	int vertexCount=0;
	boolean haveColors = false;
	Coordinate coordNode = null;
	Color colorNode = null;

    public PointSet(Loader loader) {
	super(loader);
	color = new SFNode(null);
	coord = new SFNode(null);
	initFields();
    }
    PointSet(Loader loader, SFNode coord, SFNode color) {
	super(loader);
	this.coord = coord;
	this.color = color;
	initFields();
    }
    public Object clone() {
	return new PointSet( loader, (SFNode)coord.clone(),
	    (SFNode)color.clone()); 
    }
    public javax.media.j3d.BoundingBox getBoundingBox() {
	return bounds;
    }
    public javax.media.j3d.Geometry getImplGeom() {
	return (javax.media.j3d.Geometry)impl;
    }
    public int getNumTris() {
	return 0;
    }
    public String getType() { return "PointSet"; }
    public boolean haveTexture() {
	return false;
    }
    void initFields() {
	coord.init( this, FieldSpec, Field.EXPOSED_FIELD, "coord" );
	color.init( this, FieldSpec, Field.EXPOSED_FIELD, "color" );
    }
    public void initImpl() {
	if ((coord.node != null) && (coord.node instanceof Coordinate)) {
	    coordNode = (Coordinate) coord.node;
	    if (coordNode.point.size >= 3) { 
		vertexFormat = javax.media.j3d.GeometryArray.COORDINATES;
		vertexCount = coordNode.point.size / 3;

		if ((color.node != null) && (color.node instanceof Color)) {
		    colorNode = (Color) color.node;
		    if ((colorNode.color.vals.length / 3) == vertexCount) {
			haveColors = true;
			vertexFormat |= javax.media.j3d.GeometryArray.COLOR_3;
		    }
		}
	
		impl = new PointArray(vertexCount, vertexFormat);
		impl.setCoordinates(0, coordNode.point.value, 0, vertexCount);

		if (haveColors) {
		    impl.setColors(0, colorNode.color.vals);
		}

		// else make sure j3d uses emmissive color from material
		// or track the material in the this shape

		bounds = coordNode.point.getBoundingBox();
	    }
	}
	implObject = impl;
	implReady = true;
    }
    public void notifyMethod(String eventInName, double time) {
	if(eventInName.equals("color")||eventInName.equals("coord"))
	    initImpl();
    }
}
