package com.sun.j3d.loaders.vrml97.impl;

/*
 *      %Z%%M% %I% %E% %U%
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

import java.util.Observer;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/* Implementation notes:

   The Proto is stored as a non-impl'd VRML graph (i.e. no J3D nodes) and a
   set of routes for IS maps and a set of proto internal routes.  When the
   proto is used, a ProtoInstance is created to hold the instance.  The
   instance is filled in by cloning the definintion VRML graph.  DEF'd nodes
   are mapped between the definition and the instance so that the instance
   has the same DEF names.  IS maps are set up by making routes between the
   external fields and the instance node's fields.  IS maps don't actually
   share the fields since VRML allows more than one internal field to be
   IS mapped to the external field.
*/

public class Proto implements Namespace {

	String	name;
	Loader	loader;
	Vector	nodes = new Vector();
	public Hashtable FieldSpec = new Hashtable(4);
	Hashtable 	defTable = new Hashtable();
	Vector	routes = new Vector();
	Vector	isRoutes = new Vector();

	ProtoInstance newInstance;

    public Proto(Loader loader, String initName) {
	this.loader = loader;
	name = initName;
	if(loader.debug)System.out.println("PROTO "+initName);
    }
    // Stores a new top-level node in the definition
    public void addObject(BaseNode node) {
	nodes.addElement(node);
    }
    // Stores proto-internal ROUTEs. These are stored using the node names
    // which are converted to Nodes when the ROUTE is applied.
    void addRoute(String fromName, String fromEventOut,
            String toName, String toEventIn) {
	routes.addElement(new NameRoute(fromName, fromEventOut, toName,
			toEventIn));
    }
    // PROTO definition has it's own DEF/USE namespace, managed through these
    // calls which implement the Namespace interface  (these are public because
    // interface methods must be public)
    public void define(String defName, BaseNode node) {
        defTable.put(defName, node);
    }
    String getName() {
	return name;
    }
	// generate a copy of the proto, resolving linkages to the new instance.
	// Note that the defininion has not been initImpl()'d and the new instance
	// will get initImpl() as if it were a normal node by the Parser (not here)
	ProtoInstance instance() {

	// create a new Proto and register it with the loader
	newInstance = new ProtoInstance(loader, name, this);
	if (loader.debug)
	    System.out.println(name+" Proto.instance() begins newInstance is " +
		newInstance.toStringId());
	// Tell the loader that the def/use namespace belongs to the new
	// instance
	loader.beginProtoInstance(newInstance);


	// clone the trees attached to the Proto.  Def'd nodes will
	// register with the the new instance (see also SFNode.clone() and
	// MFNode.clone()).  Use'd nodes will lookup in the new instance (see
	// ProtoUseNode.clone()

	BaseNode[] children = new BaseNode[nodes.size()];
	for (int i = 0; i < nodes.size(); i++) {
	    BaseNode baseNode = (BaseNode) nodes.elementAt(i);
	    //System.out.println(baseNode.getType());

	    if (loader.debug) {
	        System.out.println("Proto.instance() child " + i +
		    " " + baseNode);
	    }

	    // since the sensor will be in an artificial group,
	    // the group above may want the pickability information
	    if ( baseNode instanceof VrmlSensor ) {
		newInstance.containsSensor=true;
		if ( newInstance.sensors == null )
		     newInstance.sensors = new java.util.Vector();
		newInstance.sensors.addElement(baseNode);
	    }

	    BaseNode clone = (BaseNode)baseNode.clone();

	    if (i == 0) {
		newInstance.instanceNode = clone;
	    }

	    loader.registerClone(baseNode, clone);
	}

	// set up the new fields, this is done after cloning the internals
	// so that nodes in the fields can refer to the new instance nodes.
	newInstance.initFields();

	newInstance.applyIsRoutes();
	if (loader.debug) {
	    System.out.println("Proto.instance() returns instance with " +
		"VRML tree:\n" +
		newInstance.instanceNode);
	}
		loader.endProtoInstance();
	return newInstance;
	}
	// Adds an EventIn to the definition
	public void setEventIn(String eventInName, Field f) {
	//String fieldName = Field.baseName(eventInName);
		//It is incorrect to trim the basename, especially when one follows the
		// convention of having seperate set_position and position_changed fields
		// in a PROTO definition.
		f.init(null, FieldSpec, Field.EVENT_IN, eventInName);
	}
	// Adds an EventOut to the definition
	void setEventOut(String eventOutName, Field f) {
		//String fieldName = Field.baseName(eventOutName);
		//It is not correct to trim the set or changed off of a field name when one
		// can reasonably expect to have seperate fields for those.
		f.init(null, FieldSpec, Field.EVENT_OUT, eventOutName);
	}
	// Adds an exposed Field to the defintion
	public void setExposedField(String fieldName, Field f) {
		//fieldName = Field.baseName(fieldName);
		//if (loader.debug) {
	    // System.out.println(
		//"Proto.setExposedField(): adding field named \"" + fieldName +
		//"\"");
		//}
		f.init(null, FieldSpec, Field.EXPOSED_FIELD, fieldName);
	}
	// Adds an Field to the definition
	public void setField(String fieldName, Field f) {
	if (loader.debug) {
	    System.out.println("Proto.setField(): adding field named " +
	    fieldName);
	}
	f.init(null, FieldSpec, Field.FIELD, fieldName);
	}
	// sets up ROUTEs between the PROTO's external fields and the definition
	// node's fields.  At instance time the node references will be updated
	// to the instance's nodes.
	boolean setupIsMap(Field nodeField, String protoFieldName) {
	if (loader.debug) {
	    System.out.println("Proto.setupIsMap(" + nodeField.fieldName + ", " +
		protoFieldName + ");");
	}
	Field 	protoField = (Field)FieldSpec.get(protoFieldName);
	if (protoField == null) {
	    System.err.println("Proto.setupIsMap(): Can't find proto field \""+
		protoFieldName + "\"");
	    return false;
	}
	nodeField.update(protoField); // set the initial value

	// Add "is" routes for each way the proto field implies (timestamps
	// will avoid loops)
	// Note: external connections use "null" for node name
	// ProtoInstance.applyRoutes() will update these references to
	// point to the new instance
	if (protoField.isEventIn()) {
	    if (nodeField.isEventIn()) {
		isRoutes.addElement(
		    new Route(null, protoFieldName,
			      nodeField.ownerNode, nodeField.fieldName));
	    } else {
		return false;
	    }
	}
	if (protoField.isEventOut()) {
	    if (nodeField.isEventOut()) {
		isRoutes.addElement(
		    new Route(nodeField.ownerNode, nodeField.fieldName,
			      null, protoFieldName));
	    } else {
		return false;
	    }
	}
	// If the proto field is a plain field, build the EventIn route
	// to map the initial value from the parse.  There won't
	// be any further updates since you can't route to the field
	if (!protoField.isEventIn() && !protoField.isEventOut()) {
	    isRoutes.addElement(new Route(null, protoFieldName,
			    nodeField.ownerNode, nodeField.fieldName));
	}
	return true;

	}
    // prints out a good description of the PROTO, but not a parsable
    // version (IS maps aren't in the right format)
    public String toString() {
	String retval = "Proto " + name + " [\n";
	for (Enumeration e = FieldSpec.keys(); e.hasMoreElements();) {
	    String fieldName = (String)e.nextElement();
	    Field  field = (Field)FieldSpec.get(fieldName);
	    retval += "field ";
	    String fieldType = field.getClass().getName() + " ";
	    int start = fieldType.lastIndexOf(".") + 1;
	    int end = fieldType.length();
	    fieldType = fieldType.substring(start, end);
	    retval += fieldType + " " + fieldName + " " + field;
	}
	retval += "]\n{\n";
	for (int i = 0; i < nodes.size(); i++) {
	    BaseNode baseNode = (BaseNode) nodes.elementAt(i);
	    retval += baseNode + "\n";
	}
	if (isRoutes.size() > 0) {
	    retval += "# IS routes \n";
	    for (int i = 0; i < isRoutes.size(); i++) {
		Route rt = (Route) isRoutes.elementAt(i);
		retval += "# " + rt + "\n";
	    }
	}
	if (routes.size() > 0) {
	    retval += "# internal routes\n";
	    for (int i = 0; i < routes.size(); i++) {
		NameRoute rt = (NameRoute) routes.elementAt(i);
		retval += rt + "\n";
	    }
	}
	retval += "}\n";
	return retval;
    }
    public BaseNode use(String defName) {
	BaseNode orgNode = (BaseNode) defTable.get(defName);
        return new ProtoUseNode(loader, this, orgNode);
    }
}
