package com.sun.j3d.loaders.vrml97.impl;

/*
 *      %Z%%M% %I% %E% %U%
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 *
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 */

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/*
  This is the an instance of a Proto.  It is a copy of the definition, with
  the names, routes and IS maps of the defintion mapped to the instance.
  Proto.instance() will create the instance.  ProtoInstance.initImpl() will
  initalize the J3D scene graph.
*/

public class ProtoInstance extends Node implements Namespace {

	String	name;
	Proto	proto;
	Hashtable 	defTable = new Hashtable();
	BaseNode    instanceNode;
	Vector	nodesToInit = new Vector();
	Hashtable	fieldMapping = new Hashtable();
	Hashtable   nodeMapping = new Hashtable();
	// need to get the updateParent() correct
	boolean 	containsSensor = false;
	Vector 	sensors;

    ProtoInstance(Loader loader, String name, Proto proto) {
	super(loader);
	this.name = name;
	this.proto = proto;
    }
	// Add a field to the instance. The fieldMapping table is used to map
	// in case a field has multiple event names.  Node fields must be handled
	// carefully, since we want a shallow copy with the definition nodes
	// mapped to the instance nodes.
	Field addField(Field field) {
	Field newField;
	if (!fieldMapping.containsKey(field)) {
	    if (field instanceof MFNode) {
		MFNode mfField = (MFNode) field;
		newField = mfField.cloneMap(this);
		if (loader.debug) {
		    System.out.println("clone map of MFNode");
		    System.out.println("old = " + mfField.toString());
		    System.out.println("new = " + newField.toString());
		}
	    } else if (field instanceof SFNode) {
		SFNode sfField = (SFNode) field;
		newField = sfField.cloneMap(this);
		if (loader.debug) {
		    System.out.println("clone map of SFNode");
		    System.out.println("old = " + sfField.toString());
		    System.out.println("new = " + newField.toString());
		}
	    } else {
		newField = (Field) field.clone();
	    }
	    fieldMapping.put(field, newField);
	    newField.init(this, FieldSpec, field.fieldType, field.fieldName);
	} else {
	    newField = (Field)fieldMapping.get(field);
	}
	return newField;
	}
// Setup routes for IS maps in the PROTO definition
void applyIsRoutes() {
    // apply the routes, updating the nodes
    if (loader.debug) {
        System.out.println("ProtoInstance: applying IS routes:");
    }
    // apply the IS routes
    for (Enumeration e = proto.isRoutes.elements(); e.hasMoreElements();) {
        Route route = (Route) e.nextElement();
        BaseNode fromNode, toNode;

        if (route.fromNode == null) {
            fromNode = this;
        } else {
            fromNode = route.fromNode;
            if (fromNode instanceof ProtoUseNode) {
                fromNode = ((ProtoUseNode) fromNode).orgNode;
            }
            fromNode = updateNode(fromNode);
        }

        // lookup and update the toNode
        if (route.toNode == null) {
            toNode = this;
        } else {
            toNode = route.toNode;
            if (toNode instanceof ProtoUseNode) {
                toNode = ((ProtoUseNode) toNode).orgNode;
            }
            toNode = updateNode(toNode);
        }
        if (loader.debug) {
            System.out.println(
                "Add IS route: "
                    + fromNode.toStringId()
                    + " . "
                    + route.fromField
                    + "\n   to "
                    + toNode.toStringId()
                    + " . "
                    + route.toField); 
        }

        if (!route.fromField.equals("?") && !route.toField.equals("?")) {
            loader.connect(fromNode, route.fromField, toNode, route.toField, false);
        } else {
            // bug: after fixing 4221802 the multinoded protodef somehow
            // a "?" field was trying to get addRouted
            System.err.println(
                "ProtoInstance.applyIsRoutes(): " + " a field name is \"?\""); 
        }
    }
}
    // Apply the internal routes for the PROTO.
    void applyRoutes() {
	// apply the routes, updating the nodes
	if (loader.debug) {
	    System.out.println("ProtoInstance: applying routes:");
	}
	// apply the internal routes
	for (Enumeration e = proto.routes.elements(); e.hasMoreElements(); ) {
	    NameRoute route = (NameRoute) e.nextElement();
	    BaseNode fromNode, toNode;

	    // lookup and update the fromNode
	    fromNode = proto.use(route.fromName);
	    if (fromNode instanceof ProtoUseNode) {
		fromNode = ((ProtoUseNode)fromNode).orgNode;
	    }
	    fromNode = updateNode(fromNode);

	    // lookup and update the toNode
	    toNode = proto.use(route.toName);
	    if (toNode instanceof ProtoUseNode) {
		toNode = ((ProtoUseNode)toNode).orgNode;
	    }
	    toNode = updateNode(toNode);
	    if (loader.debug) {
		System.out.println("Add route: " + fromNode + " . " +
		    route.fromField + "\n   to " + toNode + " . " +
		    route.toField);
	    }
	    if ( !route.fromField.equals("?") &&
		 !route.toField.equals("?") ) {
	        loader.connect(fromNode, route.fromField, toNode,
		    route.toField, false);
	    } else {
		// bug: after fixing 4221802 the multinoded protodef somehow
		// a "?" field was trying to get addRouted
		System.err.println("ProtoInstance.applyRoutes(): " +
			" a field name is \"?\"");
	    }
	}
    }
    // This is not implemented, but is probably legal, since the a proto
    // instance could be part of another PROTO's definition, which would
    // lead to a clone() when the second PROTO is instanciated.
    public Object clone() {
	System.err.println("Warning: clone called on a ProtoInstance!");
	return null;
    }
    // While the instance is being created, the loader DEF/USE namespace
    // belongs to the instance.
    public void define(String defName, BaseNode node) {
	//System.out.println("ProtoInstance.define() name: " + defName +
	//	"value: " + node);
        defTable.put(defName, node);
    }

	/* Note that the implNode is actually just the root of the graph for this object, and not
	 * necessarily connected with the interesting mechanics of this object.  E.g. Transform objects.
	 */
	public javax.media.j3d.Node getImplNode() {
		return instanceNode.getImplNode();
	}
    // Get the J3D node for the instance node.  This is the J3D connection
    // point for the instance.
    public javax.media.j3d.SceneGraphObject getImplObject() {
	javax.media.j3d.SceneGraphObject implObject =
		instanceNode.getImplObject();
	if (loader.debug) {
	    System.out.println("ProtoInstance.getImplObject returning: ");
	    if (implObject == null) {
		System.out.println("implObject = null");
	    } else {
		loader.treePrinter.print(implObject);
	    }
	}
	return implObject;
    }
    // Returns the PROTO name
    public String getType() {
	return name;
    }
    // Setup the fields for the instance.
    void initFields() {
        // clone the Proto's fields and event mappings
	// handle the fields first
        for (Enumeration e = proto.FieldSpec.keys(); e.hasMoreElements();) {
            String fieldName = (String)e.nextElement();
            Field  field = (Field)proto.FieldSpec.get(fieldName);
	    //System.err.println("ProtoInstance.initFields(): got field " +
	    //  "named " + fieldName + " with value " + field);
	    Field newField;
	    if ((field instanceof ConstField)) {
		newField = addField(((ConstField)field).ownerField);
		FieldSpec.put(fieldName, newField.constify());
	    } else {
		newField = addField(field);
		FieldSpec.put(fieldName, newField);
	    }
        }
    }
	// This creates the J3D tree from the instance.  The nodes of the
	// tree are initImpl'd in the same order as they were added to the
	// definition.  Routes are applied after the nodes are initImpl'd to
	// allow the capibility bits to be set correctly.
	public void initImpl() {
	if (loader.debug) {
	    System.out.println("ProtoInstance.initImpl():");
	}
	for (Enumeration e = nodesToInit.elements(); e.hasMoreElements(); ) {
	    BaseNode node = (BaseNode) e.nextElement();
	    if (loader.debug) {
		System.out.println("calling initImpl on " + node.toStringId());
	    }
	    node.initImpl();
	}
	applyRoutes();
	if (loader.debug) {
	    System.out.println("instanceNode.getImplObject() = ");
	    loader.treePrinter.print(instanceNode.getImplObject());
	}
	implReady = true;

	// Technically, at this point the Proto is dead weight which could just as easily be collected.
	proto=null;

	
	}
    // Returns the first base node of the instance, which is the "type" of the
    // proto
    BaseNode instanceNode() {
	return instanceNode;
    }
    // Handle runtime updates to the instance.  Routes events pass the
    // mark the connected fields as being writable. Otherwise the ROUTEs
    // will pass the event to the node's fields, which will cause the internal
    // nodes to be updated.
    public void notifyMethod(String eventInName, double time) {
	if (eventInName.startsWith("route_")) {
	    //System.out.println("ProtoInstance: handling " + eventInName);
	    eventInName = eventInName.substring(6);
	    Field field = getField(eventInName);
	    //System.out.println("making fields connected to  " + field +
	    //	" writable");
	    field.markConnectedWriteable();
	}
    }
    // When a definition node gets cloned, this method stores the mapping from
    // the defintion node to the instance node
    void registerClone(BaseNode org, BaseNode clone) {
	//System.out.println("registerClone org = " + org.toStringId() +
        //		" clone = " + clone.toStringId());
	nodeMapping.put(org, clone);
	nodesToInit.add(clone);
    }
    public String toStringBody() {
	return "PROTO " + name + " {\n" + /* TODO: fields here */ "}\n";
    }
	// Given a node from the definition, return the corresponding node in
	// the instance.
	BaseNode updateNode(BaseNode node) {
	BaseNode retval;
	// Map internal connections using the nodeMapping table
	retval = (BaseNode)nodeMapping.get(node);
	if (loader.debug) {
	    System.out.println("ProtoInstance.updateNode() mapped "+node);
	    System.out.println(" to "+retval);
	}
	if (retval == null) {
	    System.err.println("ProtoInstance: Unable to update node " +
		"reference: " + node.toStringId());
	}
	return retval;
	}
/**
 * Inform the node that the scene graph above it has changed.
 * TimeSensors and other nodes the Browser object should know about should
 * respond by performing the appropriate registrations.
 */
public void updateSceneGraph(BaseNode parent) {
	if (instanceNode != null)
		instanceNode.updateSceneGraph(parent);
}
    public BaseNode use(String defName) {
	BaseNode retval = (BaseNode) defTable.get(defName);
	//System.out.println("ProtoInstance.use(" + defName + ") " +
	//	" returned " + " node " + retval.toStringId());
	retval.registerUse(loader.scene);
	return retval;
    }
}
