package com.sun.j3d.loaders.vrml97.impl;

import javax.vecmath.*;
/*
 *      @(#)ProximityBehavior.java 1.7 98/11/05 20:34:56
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */
/*
 * @Author: Rick Goldberg
 *
 */

import javax.media.j3d.*;


public class ProximityBehavior extends Behavior {

	ProximitySensor owner;
	WakeupOnViewPlatformEntry woven;
	WakeupOnElapsedFrames prox_gen;
	WakeupOnViewPlatformExit wovex;
	WakeupCriterion[] conditions;
	WakeupOr wor;
	boolean active = false;

	public ProximityBehavior(ProximitySensor ps) {
		setCapability(ALLOW_LOCAL_TO_VWORLD_READ);
		owner = ps;
	}
	public void initialize() { 
	setSchedulingBounds(owner.bbox);
	wovex = new WakeupOnViewPlatformExit(owner.bbox);
	woven = new WakeupOnViewPlatformEntry(owner.bbox);
	prox_gen = new WakeupOnElapsedFrames(0); // this should be fine with the bounds set.
	conditions = new WakeupCriterion[3];
	conditions[0] = wovex;
	conditions[1] = woven;
	conditions[2] = prox_gen;
	wor = new WakeupOr(conditions);
	wakeupOn(wor);
	}
// TODO: the Browser should maintain a list of objects interested
// in the current position and orientation of the view platform.
// when woven happens, we should add to that to generate 
// the proper eventOuts from the ProximitySensor;
// Additionally to correctly implement, the eventOuts must be in the
// Bounds own coordinate system.

public void processStimulus(java.util.Enumeration ofElements) {
	WakeupCriterion wakeup;
	double t = Time.getNow();
	//if (!owner.enabled.getValue())
	//  return;
	while (ofElements.hasMoreElements()) {
		wakeup = (WakeupCriterion) ofElements.nextElement();
		if (wakeup instanceof WakeupOnViewPlatformExit) {
			owner.exitTime.setValue(t);
			active = false;
			owner.isActive.setValue(active);
			if (owner.browser!=null)
				owner.browser.removeProximitySensor(this);
			else
				System.out.println("Deferring removal.");
//            wakeupOn(new WakeupOnViewPlatformEntry(owner.bbox));
			System.out.println("DEBUG: ProximityBehavior:exit");
		} else if ((wakeup instanceof WakeupOnElapsedFrames) || (wakeup instanceof WakeupOnViewPlatformEntry)) {
			owner.enterTime.setValue(t);
			active = true;
			owner.isActive.setValue(active);
			if (owner.browser!=null)
				owner.browser.addProximitySensor(this);
			else
				System.out.println("Deferring add.");
 		//              wakeupOn(new WakeupOnViewPlatformExit(owner.bbox));
			System.out.println("DEBUG: ProximityBehavior:entry");
		}
	}
	// Construct the new wake up conditions:
	WakeupCriterion wcr[]=new WakeupCriterion[2];
	wcr[0]=new WakeupOnViewPlatformEntry(owner.bbox);
	wcr[1]=new WakeupOnViewPlatformExit(owner.bbox);
	WakeupOr wor=new WakeupOr(wcr);
	wakeupOn(wor);	
}
	public void resetSchedulingBounds() {
		setSchedulingBounds(owner.bbox);
	}
	public void updateUserPosition() {
		if (true)
			return;
		
			//owner.monitor.pulse();			
			Transform3D currentView=null;
			//Transform3D currentView=owner.browser.evagation.currentView;

			Transform3D ownerToVworld=new Transform3D();
			try {
			owner.getImplNode().getLocalToVworld(ownerToVworld);
			} catch (CapabilityNotSetException e) {
				System.err.println("Error getting proximity sensor LocalToVWorld.");
				e.printStackTrace(System.err);
			} catch (RestrictedAccessException e) {
				// We are not in the scene graph right now, check back later.
				return;
			}
			Transform3D platformToVworld=new Transform3D();
			try {
//			owner.browser.curViewPlatform.getLocalToVworld(platformToVworld);
			} catch (CapabilityNotSetException e) {
				System.err.println("Error getting viewplatform LocalToVWorld.");
				e.printStackTrace(System.err);
			} catch (RestrictedAccessException e) {
				// The view isn't in the scene graph right now.  This is probably a problem.
				System.err.println("View outside of graph.");
				return;
			}
			// This is just plain wrong, since we're not checking the relative scene transforms.
			Transform3D where=new Transform3D(currentView);
			where.mul(platformToVworld);
			ownerToVworld.invert();
			where.mul(ownerToVworld);
			javax.vecmath.AxisAngle4f cAxis=new javax.vecmath.AxisAngle4f();
			Matrix3f cMatrix=new javax.vecmath.Matrix3f();
			Vector3f cPosition=new javax.vecmath.Vector3f();
			where.get(cMatrix,cPosition);
			cAxis.set(cMatrix);
			
			if (owner.browser.debug) {  
				System.out.println("User position now at ("+cPosition.x+","+cPosition.y+","+cPosition.z+")");
				System.out.println("User orientation now at ("+cAxis.x+","+cAxis.y+","+cAxis.z+","+cAxis.angle+")");
			}
			if (Float.isNaN(cAxis.x) || Float.isNaN(cAxis.y) || Float.isNaN(cAxis.z) ) {
			  cAxis.x=0.0f;
			  cAxis.y=0.0f;
			  cAxis.z=1.0f;
			}
				
			owner.position.setValue(cPosition.x, cPosition.y, cPosition.z);
			owner.orientation.setValue(cAxis.x, cAxis.y, cAxis.z, cAxis.angle);

	}
}
