package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)ProximitySensor.java 1.16 98/11/05 20:34:57
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */
/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 *
 */

import javax.media.j3d.BoundingBox;
import javax.media.j3d.WakeupOr;
import javax.media.j3d.WakeupAnd;
import javax.media.j3d.WakeupOnViewPlatformEntry;
import javax.media.j3d.WakeupOnViewPlatformExit;
import javax.vecmath.Point3d;

public class ProximitySensor extends Node {

 //edu.nodak.ndsu.GAMES.util.HeartbeatMonitor monitor;
	
	// exposedFields
	SFVec3f center;
	SFVec3f size;
	SFBool enabled;

	// eventOuts
	SFBool isActive;
	SFVec3f position;
	SFRotation orientation;
	SFTime enterTime;
	SFTime exitTime;

	// 
	javax.media.j3d.BoundingBox bbox;
	ProximityBehavior behavior;

	public ProximitySensor(Loader loader) {
	super(loader);
	//monitor=new edu.nodak.ndsu.GAMES.util.HeartbeatMonitor("PS",null);
	center = new SFVec3f(0.0f,0.0f,0.0f);
	size = new SFVec3f(0.0f,0.0f,0.0f);
	enabled = new SFBool(true);
	isActive = new SFBool(false);
	position = new SFVec3f();
	orientation = new SFRotation();
	enterTime = new SFTime();
	exitTime = new SFTime();
	initFields();
	}
	ProximitySensor(Loader loader, SFVec3f center, SFVec3f size, 
		SFBool enabled) {
	super(loader);
	//monitor=new edu.nodak.ndsu.GAMES.util.HeartbeatMonitor("PSClone",null);
	this.center = center;
	this.size = size;
	this.enabled = enabled;
	isActive = new SFBool(false);
	position = new SFVec3f();
	orientation = new SFRotation();
	enterTime = new SFTime();
	exitTime = new SFTime();
	initFields();
	}
    public Object clone() {
	return new ProximitySensor(loader,(SFVec3f)center.clone(),
	    (SFVec3f)size.clone(),(SFBool)enabled.clone());
    }
    public String getType() { return "ProximitySensor"; }
	void initFields() {
	enabled.init(	this, FieldSpec, Field.EXPOSED_FIELD, "enabled" );
	center.init( this, FieldSpec, Field.EXPOSED_FIELD, "center" );
	size.init( this, FieldSpec, Field.EXPOSED_FIELD, "size" );
	isActive.init( this, FieldSpec, Field.EVENT_OUT, "isActive" );
	position.init( this, FieldSpec, Field.EVENT_OUT, "position_changed" );
	orientation.init( this, FieldSpec, Field.EVENT_OUT, "orientation_changed" );
	enterTime.init( this, FieldSpec, Field.EVENT_OUT, "enterTime" );
	exitTime.init( this, FieldSpec, Field.EVENT_OUT, "exitTime" );
	}
	public void initImpl() {
	
	bbox = new BoundingBox( new Point3d( center.value[0] - size.value[0]/2,
					     center.value[1] - size.value[1]/2,
					     center.value[2] - size.value[2]/2 ) ,
			        new Point3d( center.value[0] + size.value[0]/2,
					     center.value[1] + size.value[1]/2,
					     center.value[2] + size.value[2]/2 ) );

	behavior = new ProximityBehavior(this);

	javax.media.j3d.BranchGroup aBranchGroup=new javax.media.j3d.BranchGroup();
	aBranchGroup.addChild(behavior);
	aBranchGroup.setCapability(javax.media.j3d.BranchGroup.ALLOW_DETACH);
	aBranchGroup.setCapability(javax.media.j3d.BranchGroup.ALLOW_LOCAL_TO_VWORLD_READ);
	implObject=aBranchGroup;
	
	implReady = true;
	}
	public void notifyMethod(String eventInName, double time) {
	if(eventInName.equals("enabled")) {
	    if(enabled.getValue() == true) {
		behavior.initialize();
	    } //else {
		//impl.conditions[0] = impl.wovex;
		//impl.conditions[1] = impl.woven;
		//// should never happen
		//impl.wakeupOn(new WakeupAnd(impl.conditions));
	    //}
	}
	}
/**
 * Inform the node that the scene graph above it has changed.
 * TimeSensors and other nodes the Browser object should know about should
 * respond by performing the appropriate registrations.
 */
public void updateSceneGraph(BaseNode newParent) {
	BaseNode oldParent=parent;
	super.updateSceneGraph(newParent);
	if (parent==null) {
		if ((browser!=null)&&(isActive.getValue())) {
			System.out.println("Removing active sensor from scenegraph.");
			browser.removeProximitySensor(behavior);
		}
	} else {
		if ((browser!=null)&&(isActive.getValue())) {
			System.out.println("Belated proximity sensor registration.");
			browser.addProximitySensor(behavior);
		} else {
			System.out.println("No registration.");
			if (browser==null)
				System.out.println("Null browser.  IsActive is "+isActive.getValue());
			else
				System.out.println("OK browser.  IsActive is "+isActive.getValue());
		}
	}
}
}
