package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)SFColor.java 1.15 98/11/05 20:35:00
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 * @Author: Doug Gehringer
 * @Author: Auvo Severinkangas
 *
 */

import java.util.Observable;
import java.util.Observer;

public class SFColor extends Field implements SFNumeric {

	float[] color;

    public SFColor(float[] initColor) {
	color = new float[3];
	setValue(initColor);
    }
    public SFColor(float red, float green, float blue) {
	color = new float[3];
	color[0] = red;
	color[1] = green;
	color[2] = blue;
    }
    public synchronized Object clone() {
	return new SFColor( color );
    }
    public synchronized ConstField constify() {
	if (constField == null) {
	    constField = new ConstSFColor(this);
	}
	return constField;
    }
    public float getBlue() {
	return color[2];
    }
    public float getGreen() {
	return color[1];	
    }
    public float getRed() {
	return color[0];
    }
    public float[] getValue() {
	return color;
    }
    public void getValue(float colr[]) {
	System.arraycopy(color,0,colr,0,3);
	    
    }
    public String getX3DValue()
    {
        return color[0] + " " + color[1] + " " + color[2];
    }
    public void setBlue(float blue) {
	color[2]=blue;
	route();
    }
    public void setGreen(float green) {
	color[1]=green;
	route();
    }
    public void setRed(float red) {
	color[0]=red;
	route();
    }
    public void setValue(float[] colrs) { 
	System.arraycopy(colrs,0,color,0,3);
	route();
    }
    public void setValue(float red, float green, float blue) {
	color[0]=red;
	color[1]=green;
	color[2]=blue;

	route();
    }
    public void setValue( ConstSFColor constsfcolr ) {
	setValue( (SFColor)constsfcolr.ownerField );
	    
    }
    public void setValue( SFColor sfcolr ) {
	setValue( sfcolr.color );
		    
    }
    public String toString() {
	return color[0] + " " + color[1] + " " + color[2] + "\n";
    }
    public void update(Field field) {
	setValue((SFColor)field );
    }
    public vrml.Field wrap() {
        return new vrml.field.SFColor(this);
    }
}
