package com.sun.j3d.loaders.vrml97.impl;

/*
 *      @(#)TouchSensor.java 1.28 99/03/10 18:02:52
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * @Author: Rick Goldberg
 *
 */

import java.util.Vector;

public class TouchSensor extends Node implements VrmlSensor {

	// exposedField 
	SFBool enabled; 

	// Event outs
	SFVec3f	hitNormal;
	SFVec3f	hitPoint;
	SFVec2f	hitTexCoord;
	SFBool	isActive;
	SFBool	isOver;
	SFTime	touchTime;

	// Scene graph reference
	javax.media.j3d.Node parentImpl;

    public TouchSensor(Loader loader ) {
	super(loader);
	enabled = new SFBool(true);
	hitNormal = new SFVec3f();
	hitPoint = new SFVec3f();
	hitTexCoord = new SFVec2f();
	isActive = new SFBool(true);
	isOver = new SFBool(false);
	touchTime = new SFTime(0.0);
	loader.addTouchSensor(this);
	initFields();
    }
    TouchSensor(Loader loader, SFBool enabled) {
	super(loader);
	this.enabled = enabled;
	hitNormal = new SFVec3f();
	hitPoint = new SFVec3f();
	hitTexCoord = new SFVec2f();
	isActive = new SFBool(true);
	isOver = new SFBool(false);
	touchTime = new SFTime(0.0);
	loader.addTouchSensor(this);
	initFields();
    }
    public Object clone() {
	TouchSensor t = new TouchSensor(loader, (SFBool)enabled.clone());
	return t;
    }
    public String getType() { return "TouchSensor"; }
	void initFields() {
	enabled.init(    this, FieldSpec, Field.EXPOSED_FIELD, "enabled");
	hitNormal.init(  this, FieldSpec, Field.EVENT_OUT,     "hitNormal_changed");
	hitPoint.init(   this, FieldSpec, Field.EVENT_OUT,     "hitPoint_changed");
	hitTexCoord.init(this, FieldSpec, Field.EVENT_OUT,     "hitTexCoord_changed");
	isActive.init(   this, FieldSpec, Field.EVENT_OUT,     "isActive");
	isOver.init(     this, FieldSpec, Field.EVENT_OUT,     "isOver");
	touchTime.init(  this, FieldSpec, Field.EVENT_OUT,     "touchTime");
	}
    public void initImpl() {
	loader.addTouchSensor(this);
    }
	public boolean isEnabled() {
		return enabled.value;
	}
    public void notifyMethod(String eventInName, double time) {
        if (eventInName.equals("enabled")) {
        } else if (eventInName.equals("route_enabled")) {
	    // No-op
        } else {
            System.err.println("TouchSensor: unknown eventInName " + 
		eventInName);
        }
    }
	public void setIsActive(boolean val) {
		if (isEnabled()) {
			isActive.setValue(val);
		}
	}
	public void setIsOver(boolean val) {
		if (isEnabled())
			isOver.setValue(val);
	}
	void updateParent(javax.media.j3d.Node parentImpl) {
	if (loader.debug) {
	    System.out.println("TouchSensor.updateParent()");
	}
	if (loader.loaderMode > Loader.LOAD_STATIC) {
	    if (loader.debug) {
		System.out.println("Touch Sensor enabling PICK_REPORTING on "
			+ " parent "+ parentImpl);
	    }
	    Vector v = (Vector)(parentImpl.getUserData());
	    if (v==null) {
		v = new Vector();
		parentImpl.setUserData(v);
		if (loader.debug) {
		    System.out.println("Touch Sensor parent: " + parentImpl + 
			" had no user data, added vector:" + v); 
		}
	    } 
	    v.addElement(this);
	    try {
	    parentImpl.setCapability(
			javax.media.j3d.Node.ENABLE_PICK_REPORTING);
	    parentImpl.setCapability(javax.media.j3d.Node.ALLOW_BOUNDS_READ);
	    parentImpl.setCapability(
			javax.media.j3d.Node.ALLOW_LOCAL_TO_VWORLD_READ);
	    } catch (Exception e) {
	    }
	    parentImpl.setPickable(true);
	}
	}
/**
 * Inform the node that the scene graph above it has changed.
 * TimeSensors and other nodes the Browser object should know about should
 * respond by performing the appropriate registrations.
 */
public void updateSceneGraph(BaseNode parent) {}
}
